@@ -1,7 +1,7 @@
 /* ------------------------------------------------------------------------- */
 
 /*
- * Copyright 2007-2015 GRAHAM DUMPLETON
+ * Copyright 2007-2016 GRAHAM DUMPLETON
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -29,6 +29,8 @@
 #include "wsgi_stream.h"
 #include "wsgi_metrics.h"
 #include "wsgi_daemon.h"
+#include "wsgi_metrics.h"
+#include "wsgi_thread.h"
 
 #if APR_HAVE_UNISTD_H
 #include <unistd.h>
@@ -96,7 +98,7 @@ static PyObject *SignalIntercept_call(
             PyObject *args = NULL;
             PyObject *result = NULL;
             Py_INCREF(o);
-            log = newLogObject(NULL, APLOG_WARNING, NULL);
+            log = newLogObject(NULL, APLOG_WARNING, NULL, 0);
             args = Py_BuildValue("(OOO)", Py_None, Py_None, log);
             result = PyEval_CallObject(o, args);
             Py_XDECREF(result);
@@ -162,11 +164,6 @@ PyTypeObject SignalIntercept_Type = {
 const char *wsgi_python_path = NULL;
 const char *wsgi_python_eggs = NULL;
 
-#if APR_HAS_THREADS
-int wsgi_thread_count = 0;
-apr_threadkey_t *wsgi_thread_key;
-#endif
-
 #if PY_MAJOR_VERSION > 3 || (PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION >= 4)
 static void ShutdownInterpreter_dealloc(ShutdownInterpreterObject *self)
 {
@@ -272,7 +269,7 @@ static PyObject *ShutdownInterpreter_call(
                         PyObject *log = NULL;
                         PyObject *args = NULL;
                         Py_INCREF(o);
-                        log = newLogObject(NULL, APLOG_ERR, NULL);
+                        log = newLogObject(NULL, APLOG_ERR, NULL, 0);
                         args = Py_BuildValue("(OOOOO)", type, value,
                                              traceback, Py_None, log);
                         result = PyEval_CallObject(o, args);
@@ -519,7 +516,7 @@ InterpreterObject *newInterpreterObject(const char *name)
      * the 'pdb' module.
      */
 
-    object = newLogObject(NULL, APLOG_ERR, "stderr");
+    object = newLogObject(NULL, APLOG_ERR, "stderr", 1);
     PySys_SetObject("stderr", object);
     Py_DECREF(object);
 
@@ -532,7 +529,7 @@ InterpreterObject *newInterpreterObject(const char *name)
             Py_DECREF(object);
         }
         else {
-            object = newLogObject(NULL, APLOG_ERR, "stdout");
+            object = newLogObject(NULL, APLOG_ERR, "stdout", 1);
             PySys_SetObject("stdout", object);
             Py_DECREF(object);
         }
@@ -1166,6 +1163,14 @@ InterpreterObject *newInterpreterObject(const char *name)
     PyModule_AddObject(module, "process_metrics", PyCFunction_New(
                        &wsgi_process_metrics_method[0], NULL));
 
+    PyModule_AddObject(module, "subscribe_events", PyCFunction_New(
+                       &wsgi_process_events_method[0], NULL));
+
+    PyModule_AddObject(module, "event_callbacks", PyList_New(0));
+
+    PyModule_AddObject(module, "request_data", PyCFunction_New(
+                       &wsgi_request_data_method[0], NULL));
+
     /* Done with the 'mod_wsgi' module. */
 
     Py_DECREF(module);
@@ -1399,32 +1404,21 @@ InterpreterObject *newInterpreterObject(const char *name)
 
     if (self->owner) {
 #if APR_HAS_THREADS
-        int thread_id = 0;
-        int *thread_handle = NULL;
+        WSGIThreadInfo *thread_handle = NULL;
 
         self->tstate_table = apr_hash_make(wsgi_server->process->pool);
 
-        apr_threadkey_private_get((void**)&thread_handle, wsgi_thread_key);
-
-        if (!thread_handle) {
-            thread_id = wsgi_thread_count++;
-            thread_handle = (int*)apr_pmemdup(wsgi_server->process->pool,
-                                              &thread_id, sizeof(thread_id));
-            apr_threadkey_private_set(thread_handle, wsgi_thread_key);
-        }
-        else {
-            thread_id = *thread_handle;
-        }
+        thread_handle = wsgi_thread_info(1, 0);
 
         if (wsgi_server_config->verbose_debugging) {
             ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, wsgi_server,
                          "mod_wsgi (pid=%d): Bind thread state for "
                          "thread %d against interpreter '%s'.", getpid(),
-                         thread_id, self->name);
+                         thread_handle->thread_id, self->name);
         }
 
-        apr_hash_set(self->tstate_table, thread_handle,
-                     sizeof(*thread_handle), tstate);
+        apr_hash_set(self->tstate_table, &thread_handle->thread_id,
+                     sizeof(thread_handle->thread_id), tstate);
 
         PyThreadState_Swap(save_tstate);
 #else
@@ -1447,6 +1441,8 @@ static void Interpreter_dealloc(InterpreterObject *self)
     PyObject *exitfunc = NULL;
 #endif
 
+    PyObject *event = NULL;
+
     /*
      * We should always enter here with the Python GIL
      * held and an active thread state. This should only
@@ -1464,23 +1460,12 @@ static void Interpreter_dealloc(InterpreterObject *self)
 
     if (*self->name) {
 #if APR_HAS_THREADS
-        int thread_id = 0;
-        int *thread_handle = NULL;
+        WSGIThreadInfo *thread_handle = NULL;
 
-        apr_threadkey_private_get((void**)&thread_handle, wsgi_thread_key);
+        thread_handle = wsgi_thread_info(1, 0);
 
-        if (!thread_handle) {
-            thread_id = wsgi_thread_count++;
-            thread_handle = (int*)apr_pmemdup(wsgi_server->process->pool,
-                                              &thread_id, sizeof(thread_id));
-            apr_threadkey_private_set(thread_handle, wsgi_thread_key);
-        }
-        else {
-            thread_id = *thread_handle;
-        }
-
-        tstate = apr_hash_get(self->tstate_table, &thread_id,
-                              sizeof(thread_id));
+        tstate = apr_hash_get(self->tstate_table, &thread_handle->thread_id,
+                              sizeof(thread_handle->thread_id));
 
         if (!tstate) {
             tstate = PyThreadState_New(self->interp);
@@ -1489,11 +1474,11 @@ static void Interpreter_dealloc(InterpreterObject *self)
                 ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, wsgi_server,
                              "mod_wsgi (pid=%d): Create thread state for "
                              "thread %d against interpreter '%s'.", getpid(),
-                             thread_id, self->name);
+                             thread_handle->thread_id, self->name);
             }
 
-            apr_hash_set(self->tstate_table, thread_handle,
-                         sizeof(*thread_handle), tstate);
+            apr_hash_set(self->tstate_table, &thread_handle->thread_id,
+                         sizeof(thread_handle->thread_id), tstate);
         }
 #else
         tstate = self->tstate;
@@ -1507,6 +1492,8 @@ static void Interpreter_dealloc(InterpreterObject *self)
         PyThreadState_Swap(tstate);
     }
 
+    /* Now destroy the sub interpreter. */
+
     if (self->owner) {
         Py_BEGIN_ALLOW_THREADS
         ap_log_error(APLOG_MARK, APLOG_INFO, 0, wsgi_server,
@@ -1522,6 +1509,14 @@ static void Interpreter_dealloc(InterpreterObject *self)
         Py_END_ALLOW_THREADS
     }
 
+    /* Publish event that process is being stopped. */
+
+    event = PyDict_New();
+
+    wsgi_publish_event("process_stopping", event);
+
+    Py_DECREF(event);
+
     /*
      * Because the thread state we are using was created outside
      * of any Python code and is not the same as the Python main
@@ -1619,7 +1614,7 @@ static void Interpreter_dealloc(InterpreterObject *self)
                         PyObject *log = NULL;
                         PyObject *args = NULL;
                         Py_INCREF(o);
-                        log = newLogObject(NULL, APLOG_ERR, NULL);
+                        log = newLogObject(NULL, APLOG_ERR, NULL, 0);
                         args = Py_BuildValue("(OOOOO)", type, value,
                                              traceback, Py_None, log);
                         result = PyEval_CallObject(o, args);
@@ -1752,7 +1747,7 @@ static void Interpreter_dealloc(InterpreterObject *self)
                     PyObject *log = NULL;
                     PyObject *args = NULL;
                     Py_INCREF(o);
-                    log = newLogObject(NULL, APLOG_ERR, NULL);
+                    log = newLogObject(NULL, APLOG_ERR, NULL, 0);
                     args = Py_BuildValue("(OOOOO)", type, value,
                                          traceback, Py_None, log);
                     result = PyEval_CallObject(o, args);
@@ -2426,23 +2421,12 @@ InterpreterObject *wsgi_acquire_interpreter(const char *name)
 
     if (*name) {
 #if APR_HAS_THREADS
-        int thread_id = 0;
-        int *thread_handle = NULL;
+        WSGIThreadInfo *thread_handle = NULL;
 
-        apr_threadkey_private_get((void**)&thread_handle, wsgi_thread_key);
+        thread_handle = wsgi_thread_info(1, 0);
 
-        if (!thread_handle) {
-            thread_id = wsgi_thread_count++;
-            thread_handle = (int*)apr_pmemdup(wsgi_server->process->pool,
-                                              &thread_id, sizeof(thread_id));
-            apr_threadkey_private_set(thread_handle, wsgi_thread_key);
-        }
-        else {
-            thread_id = *thread_handle;
-        }
-
-        tstate = apr_hash_get(handle->tstate_table, &thread_id,
-                              sizeof(thread_id));
+        tstate = apr_hash_get(handle->tstate_table, &thread_handle->thread_id,
+                              sizeof(thread_handle->thread_id));
 
         if (!tstate) {
             tstate = PyThreadState_New(interp);
@@ -2451,11 +2435,11 @@ InterpreterObject *wsgi_acquire_interpreter(const char *name)
                 ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, wsgi_server,
                              "mod_wsgi (pid=%d): Create thread state for "
                              "thread %d against interpreter '%s'.", getpid(),
-                             thread_id, handle->name);
+                             thread_handle->thread_id, handle->name);
             }
 
-            apr_hash_set(handle->tstate_table, thread_handle,
-                         sizeof(*thread_handle), tstate);
+            apr_hash_set(handle->tstate_table, &thread_handle->thread_id,
+                         sizeof(thread_handle->thread_id), tstate);
         }
 #else
         tstate = handle->tstate;
