@@ -4,7 +4,7 @@
 /* ------------------------------------------------------------------------- */
 
 /*
- * Copyright 2007-2014 GRAHAM DUMPLETON
+ * Copyright 2007-2016 GRAHAM DUMPLETON
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -24,8 +24,11 @@
 #include "wsgi_python.h"
 #include "wsgi_apache.h"
 
+#include "wsgi_thread.h"
+
 /* ------------------------------------------------------------------------- */
 
+extern apr_uint64_t wsgi_total_requests;
 extern int wsgi_active_requests;
 extern int wsgi_dump_stack_traces;
 
@@ -33,11 +36,18 @@ extern apr_thread_mutex_t* wsgi_monitor_lock;
 
 extern PyMethodDef wsgi_process_metrics_method[];
 
-extern double wsgi_start_request(void);
-extern double wsgi_end_request(void);
+extern WSGIThreadInfo *wsgi_start_request(void);
+extern void wsgi_end_request(void);
 
 extern PyMethodDef wsgi_server_metrics_method[];
 
+extern long wsgi_event_subscribers(void);
+extern void wsgi_publish_event(const char *name, PyObject *event);
+
+extern PyMethodDef wsgi_process_events_method[];
+
+extern PyMethodDef wsgi_request_data_method[];
+
 /* ------------------------------------------------------------------------- */
 
 #endif
