@@ -1,7 +1,7 @@
 /* ------------------------------------------------------------------------- */
 
 /*
- * Copyright 2007-2014 GRAHAM DUMPLETON
+ * Copyright 2007-2016 GRAHAM DUMPLETON
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -34,6 +34,10 @@ const char *wsgi_daemon_group = "";
 /* Process information. */
 
 pid_t wsgi_parent_pid = 0;
+pid_t wsgi_worker_pid = 0;
+pid_t wsgi_daemon_pid = 0;
+
+apr_time_t wsgi_restart_time = 0;
 
 /* New Relic monitoring agent. */
 
@@ -119,6 +123,7 @@ WSGIServerConfig *newWSGIServerConfig(apr_pool_t *p)
     object->script_reloading = -1;
     object->error_override = -1;
     object->chunked_request = -1;
+    object->ignore_activity = -1;
 
     object->enable_sendfile = -1;
 
