@@ -4,7 +4,7 @@
 /* ------------------------------------------------------------------------- */
 
 /*
- * Copyright 2007-2014 GRAHAM DUMPLETON
+ * Copyright 2007-2016 GRAHAM DUMPLETON
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -28,8 +28,12 @@
 
 extern server_rec *wsgi_server;
 extern pid_t wsgi_parent_pid;
+extern pid_t wsgi_worker_pid;
+extern pid_t wsgi_daemon_pid;
 extern const char *wsgi_daemon_group;
 
+extern apr_time_t wsgi_restart_time;
+
 /* New Relic monitoring agent. */
 
 extern const char *wsgi_newrelic_config_file;
@@ -103,6 +107,10 @@ typedef struct {
     int error_override;
     int chunked_request;
     int map_head_to_get;
+    int ignore_activity;
+
+    apr_array_header_t *trusted_proxy_headers;
+    apr_array_header_t *trusted_proxies;
 
     int enable_sendfile;
 
